/*
 * Copyright © 2022 DATAMART LLC
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ru.datamart.kafka.postgres.reader.controller;

import io.vertx.core.http.HttpHeaders;
import io.vertx.ext.web.RoutingContext;
import lombok.extern.slf4j.Slf4j;
import lombok.val;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;
import ru.datamart.kafka.postgres.reader.model.QueryRequest;
import ru.datamart.kafka.postgres.reader.model.ReaderProcessContext;
import ru.datamart.kafka.postgres.reader.service.ReaderExecutionService;
import ru.datamart.kafka.postgres.reader.utils.LoggerContextUtils;

import static io.netty.handler.codec.http.HttpResponseStatus.OK;
import static org.springframework.util.MimeTypeUtils.APPLICATION_JSON_VALUE;

@Slf4j
@Component
public class QueryController {
    private static final String DEFAULT_MESSAGE = "Exception during read process";
    private final ReaderExecutionService readerExecutionService;

    @Autowired
    public QueryController(ReaderExecutionService readerExecutionService) {
        this.readerExecutionService = readerExecutionService;
    }

    public void query(RoutingContext routingContext) {
        try {
            val queryRequest = routingContext.getBodyAsJson().mapTo(QueryRequest.class);
            val requestId = queryRequest.getRequestId();
            LoggerContextUtils.setRequestId(requestId);
            log.info("Request[{}]: [{}]", routingContext.normalizedPath(), queryRequest);
            val readerContext = new ReaderProcessContext(routingContext, queryRequest);
            readerExecutionService.execute(readerContext)
                    .onSuccess(ignored -> respondWithSuccess(routingContext))
                    .onFailure(t -> respondWithFailure(routingContext, t));
        } catch (Exception e) {
            log.error("Exception during request prepare, request: {}", routingContext.getBodyAsString(), e);
            respondWithFailure(routingContext, e);
        }
    }

    private void respondWithFailure(RoutingContext context, Throwable t) {
        log.error("Reading process FAILED", t);
        val message = getExceptionMessage(t);
        try {
            context.response().setStatusMessage(message);
        } catch (Exception e) {
            log.warn("Could not set status message: {}, sending default", message, e);
            context.response().setStatusMessage(DEFAULT_MESSAGE);
        }
        context.fail(t);
    }

    private String getExceptionMessage(Throwable t) {
        if (t == null || t.getMessage() == null) {
            return DEFAULT_MESSAGE;
        }

        return t.getMessage().replaceAll("\r\n|\r|\n", " ");
    }

    private void respondWithSuccess(RoutingContext context) {
        log.info("Reading process SUCCESS");
        context.response()
                .putHeader(HttpHeaders.CONTENT_TYPE, APPLICATION_JSON_VALUE)
                .setStatusCode(OK.code())
                .end();
    }
}
