/*
 * Copyright © 2022 DATAMART LLC
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ru.datamart.kafka.postgres.reader.service;

import io.vertx.core.Future;
import io.vertx.pgclient.PgPool;
import io.vertx.sqlclient.Row;
import io.vertx.sqlclient.RowStream;
import io.vertx.sqlclient.SqlConnection;
import lombok.AllArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import lombok.val;
import ru.datamart.kafka.postgres.reader.configuration.properties.PostgresProperties;
import ru.datamart.kafka.postgres.reader.model.PostgresExecutorRequest;
import ru.datamart.kafka.postgres.reader.model.QueryResultItem;

import java.util.ArrayList;
import java.util.List;
import java.util.Queue;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.function.Consumer;

@Slf4j
@AllArgsConstructor
public class PostgresExecutor {
    private final PgPool pgPool;
    private final PostgresProperties postgresProperties;

    public Future<Void> execute(PostgresExecutorRequest request, Queue<RowStream<Row>> streams, Consumer<QueryResultItem> chunkHandler) {
        return pgPool.withTransaction(sqlConnection -> receiveChunks(sqlConnection, request, streams, chunkHandler));
    }

    private Future<Void> receiveChunks(SqlConnection connection,
                                       PostgresExecutorRequest request,
                                       Queue<RowStream<Row>> streams,
                                       Consumer<QueryResultItem> chunkHandler) {
        return Future.future(p -> {
            log.info("Start query execution [{}], stream number: [{}]", request.getSql(), request.getStreamNumber());
            connection.prepare(request.getSql())
                    .onSuccess(preparedStatement -> {
                        log.debug("Query prepared successfully, stream number: [{}]", request.getStreamNumber());
                        val chunkNumber = new AtomicInteger(1);
                        val dataSet = new ArrayList<List<?>>();
                        val stream = preparedStatement.createStream(postgresProperties.getFetchSize());
                        streams.add(stream);
                        stream.exceptionHandler(t -> {
                            log.error("Exception during stream read, stream number: [{}]", request.getStreamNumber(), t);
                            closeStream(dataSet, stream);
                            p.fail(t);
                        }).endHandler(v -> {
                            try {
                                log.info("Stream read success, last chunk: [{}], stream number: [{}]", chunkNumber.get(), request.getStreamNumber());
                                chunkHandler.accept(new QueryResultItem(
                                        new ArrayList<>(dataSet),
                                        chunkNumber.getAndIncrement(),
                                        true));
                                closeStream(dataSet, stream);
                                p.complete();
                            } catch (Exception e) {
                                log.error("Exception during stream end handler, stream number: [{}]", request.getStreamNumber(), e);
                                closeStream(dataSet, stream);
                                p.fail(e);
                            }
                        }).handler(row -> {
                            try {
                                val rowData = new ArrayList<>();
                                for (int i = 0; i < row.size(); i++) {
                                    rowData.add(row.getValue(i));
                                }
                                dataSet.add(rowData);
                                if (dataSet.size() == request.getChunkSize()) {
                                    log.debug("Stream read chunk [{}] filled, stream number: [{}]", chunkNumber.get(), request.getStreamNumber());
                                    chunkHandler.accept(new QueryResultItem(
                                            new ArrayList<>(dataSet),
                                            chunkNumber.getAndIncrement(),
                                            false));
                                    dataSet.clear();
                                }
                            } catch (Exception e) {
                                log.error("Exception during stream handle, stream number: [{}]", request.getStreamNumber(), e);
                                closeStream(dataSet, stream);
                                p.fail(e);
                            }
                        });
                    })
                    .onFailure(t -> {
                        log.error("Exception during query prepare, stream number: [{}]", request.getStreamNumber(), t);
                        p.fail(t);
                    });
        });
    }

    private void closeStream(ArrayList<List<?>> dataSet, RowStream<Row> stream) {
        dataSet.clear();
        stream.close();
    }
}
