/*
 * Copyright © 2022 DATAMART LLC
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ru.datamart.kafka.postgres.reader.verticle;

import io.vertx.core.CompositeFuture;
import io.vertx.core.DeploymentOptions;
import io.vertx.core.Future;
import io.vertx.sqlclient.RowStream;
import lombok.extern.slf4j.Slf4j;
import lombok.val;
import ru.datamart.kafka.postgres.reader.configuration.properties.ReaderProperties;
import ru.datamart.kafka.postgres.reader.model.Chunk;
import ru.datamart.kafka.postgres.reader.model.PostgresExecutorRequest;
import ru.datamart.kafka.postgres.reader.model.ReaderProcessContext;
import ru.datamart.kafka.postgres.reader.service.PostgresExecutor;

import java.util.ArrayList;
import java.util.List;
import java.util.stream.IntStream;

import static ru.datamart.kafka.postgres.reader.configuration.properties.VerticleProperties.QueryProperties;

@Slf4j
public class PostgresExecutorVerticle extends ConfigurableVerticle {
    private final ReaderProcessContext context;
    private final List<PostgresExecutor> databaseExecutors;
    private final ReaderProperties readerProperties;
    private final QueryProperties properties;
    private volatile long timerId = -1;

    public PostgresExecutorVerticle(ReaderProcessContext context,
                                    List<PostgresExecutor> databaseExecutors,
                                    ReaderProperties readerProperties,
                                    QueryProperties properties) {
        this.context = context;
        this.databaseExecutors = databaseExecutors;
        this.readerProperties = readerProperties;
        this.properties = properties;
    }

    @Override
    public DeploymentOptions getDeploymentOptions() {
        return new DeploymentOptions()
                .setWorker(true)
                .setWorkerPoolName("postgres-worker-pool")
                .setWorkerPoolSize(properties.getPoolSize());
    }

    @Override
    public void start() {
        vertx.eventBus().consumer(context.getStartAddress(), ignored -> execute()
                .onSuccess(event -> {
                    log.info("Query execute SUCCESS");
                    context.updateStateFinishExecution();
                })
                .onFailure(throwable -> {
                    log.error("Query execute FAILED", throwable);
                    context.updateStateSetError(throwable);
                }));
    }

    @Override
    public void stop() {
        vertx.cancelTimer(timerId);
    }

    private Future<Void> execute() {
        timerId = vertx.setPeriodic(500, event -> {
            if (context.getUploadingCount() >= readerProperties.getUploadChunkBuffer() && context.getPausedStreams().compareAndSet(false, true)) {
                context.getStreams().forEach(RowStream::pause);
                log.debug("Paused query read stream, chunk buffer filled [{}]", readerProperties.getUploadChunkBuffer());
            }

            if (context.getUploadingCount() <= readerProperties.getUploadChunkBuffer() / 10 && context.getPausedStreams().compareAndSet(true, false)) {
                context.getStreams().forEach(RowStream::resume);
                log.debug("Resumed query read stream, chunk buffer freed [{}]", readerProperties.getUploadChunkBuffer());
            }
        });

        return Future.future(promise -> {
            val queryFutures = new ArrayList<Future>(databaseExecutors.size());
            IntStream.range(0, databaseExecutors.size()).forEach(i -> {
                val queryRequest = context.getQueryRequest();
                val postgresExecutorRequest = PostgresExecutorRequest.builder()
                        .streamNumber(i)
                        .streamTotal(databaseExecutors.size())
                        .chunkSize(queryRequest.getChunkSize())
                        .sql(queryRequest.getSql())
                        .build();
                log.info("Start executor [{}] for request: [{}]", i, queryRequest.getSql());
                queryFutures.add(executeQuery(databaseExecutors.get(i), postgresExecutorRequest, context));
            });
            CompositeFuture.join(queryFutures)
                    .<Void>mapEmpty()
                    .onComplete(promise);
        });
    }


    private Future<Void> executeQuery(PostgresExecutor executor,
                                      PostgresExecutorRequest request,
                                      ReaderProcessContext context) {
        return executor.execute(request, context.getStreams(), chunk -> {
            val error = context.getState().getError();
            if (error != null) {
                throw new IllegalStateException("Stopping database stream because of process exception", error);
            }

            context.incrementUploadingCount();
            context.getChunksToUpload().add(new Chunk(chunk, request.getStreamNumber(), request.getStreamTotal()));
            vertx.eventBus().send(context.getProcessChunkAddress(), null);
        });
    }
}
