/*
 * Copyright © 2022 DATAMART LLC
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ru.datamart.kafka.postgres.reader.controller;


import io.vertx.core.Future;
import io.vertx.core.http.HttpHeaders;
import io.vertx.core.http.HttpServerResponse;
import io.vertx.core.json.JsonObject;
import io.vertx.ext.web.RoutingContext;
import lombok.val;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.extension.ExtendWith;
import org.mockito.Mock;
import org.mockito.junit.jupiter.MockitoExtension;
import ru.datamart.kafka.postgres.reader.model.QueryRequest;
import ru.datamart.kafka.postgres.reader.service.ReaderExecutionService;

import static org.mockito.ArgumentMatchers.any;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.when;
import static org.springframework.util.MimeTypeUtils.APPLICATION_JSON_VALUE;

@ExtendWith(MockitoExtension.class)
public class QueryControllerTest {

    private QueryController queryController;

    @Mock
    private RoutingContext context;

    @Mock
    private HttpServerResponse response;

    @Mock
    private ReaderExecutionService readerExecutionService;

    @Mock
    private JsonObject bodyAsJson;


    @BeforeEach
    public void setUp() {
        queryController = new QueryController(readerExecutionService);
    }

    @Test
    public void shouldFailWhenParseException() {
        // arrange
        when(context.getBodyAsJson()).thenReturn(bodyAsJson);
        when(bodyAsJson.mapTo(QueryRequest.class)).thenThrow(new RuntimeException("Exception message"));
        when(context.response()).thenReturn(response);

        // act
        queryController.query(context);

        // assert
        verify(response).setStatusMessage("Exception message");
        verify(context).fail(any());
    }

    @Test
    public void shouldFailWhenExecutionException() {
        // arrange
        val queryRequest = new QueryRequest();
        queryRequest.setRequestId("requestId");

        when(context.getBodyAsJson()).thenReturn(bodyAsJson);
        when(bodyAsJson.mapTo(QueryRequest.class)).thenReturn(queryRequest);
        when(readerExecutionService.execute(any())).thenReturn(Future.failedFuture(new Exception("Exception message")));
        when(context.response()).thenReturn(response);

        // act
        queryController.query(context);

        // assert
        verify(response).setStatusMessage("Exception message");
        verify(context).fail(any());
    }

    @Test
    public void shouldSuccess() {
        // arrange
        val queryRequest = new QueryRequest();
        queryRequest.setRequestId("requestId");

        when(context.getBodyAsJson()).thenReturn(bodyAsJson);
        when(bodyAsJson.mapTo(QueryRequest.class)).thenReturn(queryRequest);
        when(readerExecutionService.execute(any())).thenReturn(Future.succeededFuture());
        when(context.response()).thenReturn(response);

        // act
        queryController.query(context);

        // assert
        verify(response).putHeader(HttpHeaders.CONTENT_TYPE, APPLICATION_JSON_VALUE);
    }

}

