/*
 * Copyright © 2022 DATAMART LLC
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ru.datamart.kafka.postgres.reader.verticle;

import io.vertx.core.Future;
import io.vertx.core.Vertx;
import io.vertx.junit5.VertxExtension;
import io.vertx.junit5.VertxTestContext;
import lombok.val;
import org.apache.avro.Schema;
import org.junit.jupiter.api.AfterEach;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.extension.ExtendWith;
import org.mockito.Mock;
import org.mockito.junit.jupiter.MockitoExtension;
import ru.datamart.kafka.postgres.reader.configuration.properties.VerticleProperties;
import ru.datamart.kafka.postgres.reader.factory.QueryResultUpstreamFactory;
import ru.datamart.kafka.postgres.reader.model.*;
import ru.datamart.kafka.postgres.reader.upstream.QueryResultUpstream;

import java.util.Arrays;
import java.util.Collections;
import java.util.concurrent.ConcurrentLinkedQueue;

import static org.mockito.Mockito.*;

@ExtendWith({MockitoExtension.class, VertxExtension.class})
class KafkaProduceVerticleTest {
    private static final String S = "{\n" +
            "    \"type\": \"record\",\n" +
            "    \"name\": \"dwnl_test\",\n" +
            "    \"namespace\": \"pva_pxf_test\",\n" +
            "    \"fields\": [\n" +
            "      {\n" +
            "        \"name\": \"id\",\n" +
            "        \"type\": [\n" +
            "          \"null\",\n" +
            "          \"long\"\n" +
            "        ],\n" +
            "        \"default\": null\n" +
            "      }\n" +
            "    ]\n" +
            "  }";
    public static final int STREAM_ID = 0;
    public static final int STREAM_TOTAL = 1;
    public static final String TABLE_NAME = "table";
    public static final String TOPIC = "topic";
    @Mock
    private ReaderProcessContext processContext;
    @Mock
    private QueryResultUpstreamFactory queryResultUpstreamFactory;
    @Mock
    private QueryResultUpstream upstream;
    @Mock
    private VerticleProperties.QueryProperties properties;
    @Mock
    private Chunk chunk;

    private volatile String verticle;

    @BeforeEach
    void setUp(Vertx vertx, VertxTestContext testContext) {
        val queryRequest = new QueryRequest();
        queryRequest.setTable(TABLE_NAME);
        queryRequest.setKafkaTopic(TOPIC);
        queryRequest.setAvroSchema(new Schema.Parser().parse(S));
        queryRequest.setKafkaBrokers(Arrays.asList(new KafkaBrokerInfo("localhost", 1111)));

        lenient().when(processContext.getQueryRequest()).thenReturn(queryRequest);
        lenient().when(processContext.getProcessChunkAddress()).thenReturn("processchunk");
        lenient().when(queryResultUpstreamFactory.create(any(), any())).thenReturn(upstream);
        lenient().when(processContext.getState()).thenReturn(new ReaderProcessState(false, null));
        vertx.deployVerticle(new KafkaProduceVerticle(processContext, queryResultUpstreamFactory, properties))
                .onSuccess(deploymentId -> {
                    verticle = deploymentId;
                    testContext.completeNow();
                })
                .onFailure(testContext::failNow);
    }

    @AfterEach
    void cleanUp(Vertx vertx) {
        vertx.undeploy(verticle);
    }

    @Test
    void shouldSuccessWhenChunkProcessed(Vertx vertx, VertxTestContext testContext) {
        // arrange
        ConcurrentLinkedQueue<Chunk> queue = new ConcurrentLinkedQueue<>();
        queue.add(chunk);

        when(processContext.getChunksToUpload()).thenReturn(queue);
        when(upstream.push(any(), anyString(), anyString(), anyInt(), anyInt())).thenReturn(Future.succeededFuture());

        val queryResultItem = new QueryResultItem(Collections.emptyList(), 0, true);
        when(chunk.getStreamId()).thenReturn(STREAM_ID);
        when(chunk.getStreamTotal()).thenReturn(STREAM_TOTAL);
        when(chunk.getQueryResultItem()).thenReturn(queryResultItem);

        // act
        vertx.eventBus().send(processContext.getProcessChunkAddress(), null);

        // assert

        testContext.verify(() -> {
            verify(upstream, timeout(1000L).times(1)).push(same(queryResultItem), eq(TABLE_NAME), eq(TOPIC), eq(STREAM_ID), eq(STREAM_TOTAL));
            verify(processContext, timeout(1000L).times(0)).updateStateSetError(any());
            verify(processContext, timeout(1000L).times(1)).decrementUploadingCount();
        }).completeNow();
    }

    @Test
    void shouldFailWhenUpstreamPushFailed(Vertx vertx, VertxTestContext testContext) {
        // arrange
        ConcurrentLinkedQueue<Chunk> queue = new ConcurrentLinkedQueue<>();
        queue.add(chunk);

        when(processContext.getChunksToUpload()).thenReturn(queue);
        when(upstream.push(any(), anyString(), anyString(), anyInt(), anyInt())).thenReturn(Future.failedFuture(new RuntimeException("Exception")));

        val queryResultItem = new QueryResultItem(Collections.emptyList(), 0, true);
        when(chunk.getStreamId()).thenReturn(STREAM_ID);
        when(chunk.getStreamTotal()).thenReturn(STREAM_TOTAL);
        when(chunk.getQueryResultItem()).thenReturn(queryResultItem);

        // act
        vertx.eventBus().send(processContext.getProcessChunkAddress(), null);

        // assert

        testContext.verify(() -> {
            verify(upstream, timeout(1000L).times(1)).push(same(queryResultItem), eq(TABLE_NAME), eq(TOPIC), eq(STREAM_ID), eq(STREAM_TOTAL));
            verify(processContext, timeout(1000L).times(1)).updateStateSetError(any());
            verify(processContext, timeout(1000L).times(1)).decrementUploadingCount();
        }).completeNow();
    }

    @Test
    void shouldSkipWhenNoChunksInQueue(Vertx vertx, VertxTestContext testContext) {
        // arrange
        ConcurrentLinkedQueue<Chunk> queue = new ConcurrentLinkedQueue<>();
        when(processContext.getChunksToUpload()).thenReturn(queue);

        // act
        vertx.eventBus().send(processContext.getProcessChunkAddress(), null);

        // assert
        testContext.verify(() -> {
            verify(upstream, timeout(1000L).times(0)).push(any(), anyString(), anyString(), anyInt(), anyInt());
            verify(processContext, timeout(1000L).times(0)).updateStateSetError(any());
            verify(processContext, timeout(1000L).times(1)).decrementUploadingCount();
        }).completeNow();
    }

    @Test
    void shouldSkipWhenProcessFailed(Vertx vertx, VertxTestContext testContext) {
        // arrange
        when(processContext.getState()).thenReturn(new ReaderProcessState(false, new RuntimeException("Exception")));

        // act
        vertx.eventBus().send(processContext.getProcessChunkAddress(), null);

        // assert
        testContext.verify(() -> {
            verify(upstream, timeout(1000L).times(0)).push(any(), anyString(), anyString(), anyInt(), anyInt());
            verify(processContext, timeout(1000L).times(0)).updateStateSetError(any());
            verify(processContext, timeout(1000L).times(1)).decrementUploadingCount();
        }).completeNow();
    }
}