/*
 * Copyright © 2022 DATAMART LLC
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ru.datamart.kafka.postgres.reader.verticle;

import io.vertx.core.Future;
import io.vertx.core.Promise;
import io.vertx.core.Vertx;
import io.vertx.junit5.VertxExtension;
import io.vertx.junit5.VertxTestContext;
import io.vertx.sqlclient.Row;
import io.vertx.sqlclient.RowStream;
import lombok.val;
import org.junit.jupiter.api.AfterEach;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.extension.ExtendWith;
import org.mockito.Mock;
import org.mockito.junit.jupiter.MockitoExtension;
import ru.datamart.kafka.postgres.reader.configuration.properties.ReaderProperties;
import ru.datamart.kafka.postgres.reader.configuration.properties.VerticleProperties;
import ru.datamart.kafka.postgres.reader.model.QueryRequest;
import ru.datamart.kafka.postgres.reader.model.ReaderProcessContext;
import ru.datamart.kafka.postgres.reader.service.PostgresExecutor;

import java.util.Arrays;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.concurrent.atomic.AtomicBoolean;

import static org.junit.jupiter.api.Assertions.assertFalse;
import static org.junit.jupiter.api.Assertions.assertTrue;
import static org.mockito.Mockito.*;

@ExtendWith({VertxExtension.class, MockitoExtension.class})
class PostgresExecutorVerticleTest {
    public static final String START_ADDRESS = "start";
    @Mock
    private ReaderProcessContext context;
    @Mock
    private PostgresExecutor databaseExecutors1;
    @Mock
    private PostgresExecutor databaseExecutors2;
    @Mock
    private ReaderProperties readerProperties;
    @Mock
    private VerticleProperties.QueryProperties properties;
    @Mock
    private RowStream<Row> stream;

    private volatile String verticle;

    @BeforeEach
    void setUp(Vertx vertx, VertxTestContext testContext) {
        lenient().when(readerProperties.getUploadChunkBuffer()).thenReturn(1000L);
        lenient().when(context.getUploadingCount()).thenReturn(0L);
        lenient().when(context.getStartAddress()).thenReturn(START_ADDRESS);
        QueryRequest value = new QueryRequest();
        lenient().when(context.getQueryRequest()).thenReturn(value);
        val streams = new ConcurrentLinkedQueue<>(Arrays.asList(stream));
        lenient().when(context.getStreams()).thenReturn(streams);
        val pausedStreams = new AtomicBoolean(false);
        lenient().when(context.getPausedStreams()).thenReturn(pausedStreams);

        vertx.deployVerticle(new PostgresExecutorVerticle(context, Arrays.asList(databaseExecutors1, databaseExecutors2), readerProperties, properties))
                .onSuccess(deploymentId -> {
                    verticle = deploymentId;
                    testContext.completeNow();
                })
                .onFailure(testContext::failNow);
    }

    @AfterEach
    void cleanUp(Vertx vertx) {
        vertx.undeploy(verticle);
    }

    @Test
    void shouldSuccessWhenEverythingOk(VertxTestContext testContext, Vertx vertx) {
        // arrange
        when(databaseExecutors1.execute(any(), any(), any())).thenReturn(Future.succeededFuture());
        when(databaseExecutors2.execute(any(), any(), any())).thenReturn(Future.succeededFuture());

        // act
        vertx.eventBus().send(START_ADDRESS, null);

        // assert
        testContext.verify(() -> {
            verify(databaseExecutors1, timeout(1000L).times(1)).execute(any(), any(), any());
            verify(databaseExecutors2, timeout(1000L).times(1)).execute(any(), any(), any());

            verify(context, timeout(1000L).times(1)).updateStateFinishExecution();
        }).completeNow();
    }

    @Test
    void shouldStopAndResumeStreamWhenChunkBufferFilled(VertxTestContext testContext, Vertx vertx) {
        // arrange
        when(context.getUploadingCount()).thenReturn(1000L);

        Promise<Void> promise = Promise.promise();
        when(databaseExecutors1.execute(any(), any(), any())).thenReturn(promise.future());
        when(databaseExecutors2.execute(any(), any(), any())).thenReturn(promise.future());

        // act
        vertx.eventBus().send(START_ADDRESS, null);

        // assert
        testContext.verify(() -> {
            verify(stream, timeout(1000L).times(1)).pause();
            assertTrue(context.getPausedStreams().get());

            when(context.getUploadingCount()).thenReturn(0L);
            verify(stream, timeout(1000L).times(1)).resume();
            assertFalse(context.getPausedStreams().get());
        }).completeNow();
    }

    @Test
    void shouldFailWhenOneExecutionFailed(VertxTestContext testContext, Vertx vertx) {
        // arrange
        val exception = new RuntimeException("Exception");
        when(databaseExecutors1.execute(any(), any(), any())).thenReturn(Future.failedFuture(exception));
        when(databaseExecutors2.execute(any(), any(), any())).thenReturn(Future.succeededFuture());

        // act
        vertx.eventBus().send(START_ADDRESS, null);

        // assert
        testContext.verify(() -> {
            verify(databaseExecutors1, timeout(1000L).times(1)).execute(any(), any(), any());
            verify(databaseExecutors2, timeout(1000L).times(1)).execute(any(), any(), any());

            verify(context, timeout(1000L).times(1)).updateStateSetError(same(exception));
        }).completeNow();
    }
}