/*
 * Copyright © 2022 DATAMART LLC
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ru.datamart.kafka.postgres.writer.controller;

import io.vertx.core.Vertx;
import io.vertx.core.http.HttpHeaders;
import io.vertx.ext.web.RoutingContext;
import lombok.RequiredArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import lombok.val;
import org.springframework.stereotype.Component;
import org.springframework.util.MimeTypeUtils;
import ru.datamart.kafka.postgres.writer.model.*;
import ru.datamart.kafka.postgres.writer.repository.InsertDataContextRepository;
import ru.datamart.kafka.postgres.writer.service.executor.InsertDataRequestExecutor;
import ru.datamart.kafka.postgres.writer.utils.LoggerContextUtils;

import static io.netty.handler.codec.http.HttpResponseStatus.OK;

@Slf4j
@Component
@RequiredArgsConstructor
public class InsertDataController {
    private final InsertDataContextRepository dataContextRepository;
    private final InsertDataRequestExecutor executor;
    private final Vertx vertx;

    public void startLoad(RoutingContext routingContext) {
        try {
            val request = routingContext.getBodyAsJson().mapTo(InsertDataRequest.class);
            LoggerContextUtils.setRequestId(request.getRequestId());
            log.info("Request[{}]: [{}]", routingContext.normalizedPath(), request);
            val insertDataContextId = InsertDataContextId.of(request);
            val insertDataContext = dataContextRepository.getOrCreate(insertDataContextId, () -> new InsertDataContext(request, insertDataContextId));

            val insertDataContextRequestId = insertDataContext.getRequest().getRequestId();
            if (!insertDataContextRequestId.equals(request.getRequestId())) {
                log.info("Switching requestId: [{}]", insertDataContextRequestId);
                LoggerContextUtils.setRequestId(insertDataContextRequestId);
            }

            if (insertDataContext.getStarted().compareAndSet(false, true)) {
                executor.execute(insertDataContext, routingContext);
                return;
            }

            executor.resume(insertDataContext, routingContext);
        } catch (Exception e) {
            log.error("Exception during request prepare, request: {}", routingContext.getBodyAsString(), e);
            routingContext.fail(e);
        }
    }

    public void stopLoad(RoutingContext routingContext) {
        try {
            val request = routingContext.getBodyAsJson().mapTo(StopRequest.class);
            LoggerContextUtils.setRequestId(request.getRequestId());
            log.info("Request[{}]: [{}]", routingContext.normalizedPath(), request);
            val insertDataContextId = InsertDataContextId.of(request);
            dataContextRepository.get(insertDataContextId)
                    .ifPresent(insertDataContext -> {
                        val insertDataContextRequestId = insertDataContext.getRequest().getRequestId();
                        if (!insertDataContextRequestId.equals(request.getRequestId())) {
                            log.info("Switching requestId: [{}]", insertDataContextRequestId);
                            LoggerContextUtils.setRequestId(insertDataContextRequestId);
                        }
                        vertx.eventBus().send(DataTopic.SEND_RESPONSE.getValue(), insertDataContextId.getId());
                    });
            routingContext.response()
                    .putHeader(HttpHeaders.CONTENT_TYPE, MimeTypeUtils.APPLICATION_JSON_VALUE)
                    .setStatusCode(OK.code())
                    .end();
        } catch (Exception e) {
            log.error("Exception during request prepare, request: {}", routingContext.getBodyAsString(), e);
            routingContext.fail(e);
        }
    }

}
