/*
 * Copyright © 2022 DATAMART LLC
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ru.datamart.kafka.postgres.writer.service.executor;

import io.vertx.core.CompositeFuture;
import io.vertx.core.Future;
import io.vertx.core.Vertx;
import io.vertx.core.http.HttpHeaders;
import io.vertx.ext.web.RoutingContext;
import io.vertx.kafka.client.common.TopicPartition;
import lombok.AllArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import lombok.val;
import org.springframework.stereotype.Service;
import org.springframework.util.MimeTypeUtils;
import ru.datamart.kafka.postgres.writer.configuration.properties.VerticleProperties;
import ru.datamart.kafka.postgres.writer.factory.InsertRequestFactory;
import ru.datamart.kafka.postgres.writer.model.DataTopic;
import ru.datamart.kafka.postgres.writer.model.InsertDataContext;
import ru.datamart.kafka.postgres.writer.model.kafka.InsertChunk;
import ru.datamart.kafka.postgres.writer.model.kafka.TopicPartitionConsumer;
import ru.datamart.kafka.postgres.writer.service.kafka.KafkaConsumerService;
import ru.datamart.kafka.postgres.writer.verticle.ConfigurableVerticle;
import ru.datamart.kafka.postgres.writer.verticle.InsertVerticle;
import ru.datamart.kafka.postgres.writer.verticle.KafkaCommitVerticle;
import ru.datamart.kafka.postgres.writer.verticle.KafkaConsumerVerticle;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Queue;
import java.util.concurrent.ConcurrentLinkedDeque;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.stream.Collectors;

@Slf4j
@Service
@AllArgsConstructor
public class InsertDataRequestExecutor {
    private final InsertRequestFactory insertRequestFactory;
    private final VerticleProperties verticleProperties;
    private final KafkaConsumerService consumerService;
    private final List<PostgresExecutor> postgresExecutors;
    private final Vertx vertxCore;

    public void execute(InsertDataContext insertDataContext, RoutingContext routingContext) {
        insertDataContext.setInsertSql(insertRequestFactory.getSql(insertDataContext));
        runKafkaProcess(insertDataContext, routingContext);
    }

    public void resume(InsertDataContext insertDataContext, RoutingContext routingContext) {
        sendCompletedResponse(insertDataContext, routingContext);
    }

    private void runKafkaProcess(InsertDataContext insertDataContext, RoutingContext routingContext) {
        consumerService.getTopicPartitions(insertDataContext)
                .compose(topicPartitions -> {
                    val insertChunkQueue = new ConcurrentLinkedDeque<InsertChunk>();
                    val consumerMap = new HashMap<TopicPartition, TopicPartitionConsumer>();
                    val verticles = new ArrayList<ConfigurableVerticle>();

                    postgresExecutors.stream()
                            .map(executor -> InsertVerticle.builder()
                                    .executor(executor)
                                    .properties(verticleProperties.getInsert())
                                    .insertRequestFactory(insertRequestFactory)
                                    .insertChunkQueue(insertChunkQueue)
                                    .context(insertDataContext)
                                    .build())
                            .forEach(verticles::add);

                    verticles.add(
                            KafkaCommitVerticle.builder()
                                    .workerProperties(verticleProperties.getCommit())
                                    .insertRequestFactory(insertRequestFactory)
                                    .consumerService(consumerService)
                                    .consumerMap(consumerMap)
                                    .context(insertDataContext)
                                    .build()
                    );

                    topicPartitions.stream()
                            .map(topicPartition -> KafkaConsumerVerticle.builder()
                                    .workerProperties(verticleProperties.getConsumer())
                                    .insertRequestFactory(insertRequestFactory)
                                    .insertChunkQueue(insertChunkQueue)
                                    .consumerService(consumerService)
                                    .partitionInfo(topicPartition)
                                    .consumerMap(consumerMap)
                                    .context(insertDataContext)
                                    .build())
                            .forEach(verticles::add);

                    val deployedVerticals = new ConcurrentLinkedQueue<String>();
                    return CompositeFuture.join(verticles.stream()
                                    .map(kafkaConsumerVerticle -> getVerticleFuture(kafkaConsumerVerticle, deployedVerticals))
                                    .collect(Collectors.toList()))
                            .onFailure(e -> deployedVerticals.forEach(vertxCore::undeploy));
                })
                .onSuccess(success -> {
                    log.info("Loading start succeeded");
                    insertDataContext.getVerticleIds().addAll(success.list());
                    vertxCore.eventBus().publish(KafkaConsumerVerticle.START_TOPIC + insertDataContext.getId(), "");
                    vertxCore.eventBus().publish(InsertVerticle.INSERT_START_TOPIC + insertDataContext.getId(), "");
                    vertxCore.eventBus().publish(KafkaCommitVerticle.START_COMMIT + insertDataContext.getId(), "");
                    sendCompletedResponse(insertDataContext, routingContext);
                })
                .onFailure(e -> {
                    log.error("Loading start failed", e);
                    sendFailedResponse(insertDataContext, routingContext, e);
                });
    }

    private Future<String> getVerticleFuture(ConfigurableVerticle kafkaConsumerVerticle, Queue<String> deployedVerticals) {
        return vertxCore.deployVerticle(kafkaConsumerVerticle,
                        kafkaConsumerVerticle.getDeploymentOptions())
                .onSuccess(deployedVerticals::add);
    }

    private void sendCompletedResponse(InsertDataContext insertDataContext, RoutingContext routingContext) {
        try {
            routingContext.response()
                    .putHeader(HttpHeaders.CONTENT_TYPE, MimeTypeUtils.APPLICATION_JSON_VALUE)
                    .setStatusCode(200)
                    .end();
        } catch (Exception e) {
            log.error("Can't send completed response [{}]: {}", insertDataContext.getRequest().getKafkaTopic(), e.getMessage(), e);
            sendFailedResponse(insertDataContext, routingContext, e);
        }
    }

    private void sendFailedResponse(InsertDataContext insertDataContext, RoutingContext routingContext, Throwable throwable) {
        insertDataContext.setCause(throwable);
        val contextId = insertDataContext.getId();
        vertxCore.eventBus().publish(DataTopic.SEND_RESPONSE.getValue(), contextId.getId());
        routingContext.fail(throwable);
    }

}
