/*
 * Copyright © 2022 DATAMART LLC
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ru.datamart.kafka.postgres.writer.verticle;

import io.vertx.core.DeploymentOptions;
import io.vertx.core.Future;
import io.vertx.core.Promise;
import io.vertx.core.json.jackson.DatabindCodec;
import lombok.Builder;
import lombok.Getter;
import lombok.SneakyThrows;
import lombok.extern.slf4j.Slf4j;
import lombok.val;
import ru.datamart.kafka.postgres.writer.configuration.properties.VerticleProperties;
import ru.datamart.kafka.postgres.writer.factory.InsertRequestFactory;
import ru.datamart.kafka.postgres.writer.model.DataTopic;
import ru.datamart.kafka.postgres.writer.model.InsertDataContext;
import ru.datamart.kafka.postgres.writer.model.InsertDataContextId;
import ru.datamart.kafka.postgres.writer.model.kafka.InsertChunk;
import ru.datamart.kafka.postgres.writer.model.kafka.PartitionOffset;
import ru.datamart.kafka.postgres.writer.service.executor.PostgresExecutor;

import java.util.ArrayList;
import java.util.Queue;
import java.util.concurrent.atomic.AtomicReference;

@Slf4j
@Getter
@Builder
public class InsertVerticle extends ConfigurableVerticle {
    public static final String INSERT_START_TOPIC = "insert_start";
    private final VerticleProperties.InsertWorkerProperties properties;
    private final InsertRequestFactory insertRequestFactory;
    private final PostgresExecutor executor;
    private final InsertDataContext context;
    private final Queue<InsertChunk> insertChunkQueue;
    private volatile boolean stopped;
    private volatile long timerId;
    @Builder.Default
    private AtomicReference<Future<Void>> processFuture = new AtomicReference<>(Future.succeededFuture());

    @Override
    public void start() {
        vertx.eventBus().consumer(INSERT_START_TOPIC + context.getId(),
                ar -> runProcessInserts());
    }

    @Override
    public void stop(Promise<Void> promise) {
        stopped = true;
        vertx.cancelTimer(timerId);
        processFuture.get()
                .onComplete(ar -> {
                    if (ar.failed()) {
                        log.error("Process future failed on undeploy.", ar.cause());
                    }
                    promise.complete();
                });
    }

    private void runProcessInserts() {
        if (stopped) {
            return;
        }

        timerId = vertx.setTimer(properties.getInsertPeriodMs(), timer -> {
            if (stopped) {
                return;
            }

            log.debug("Batch queue size [{}], process: [{}]", insertChunkQueue.size(), InsertDataContextId.of(context.getRequest()));
            processFuture.set(processChunk()
                    .onSuccess(v -> runProcessInserts())
                    .onFailure(t -> error(context, t)));
        });
    }

    private Future<Void> processChunk() {
        return Future.future(promise -> {
            val insertChunk = insertChunkQueue.poll();
            if (insertChunk == null) {
                promise.complete();
                return;
            }
            val partitionOffsets = new ArrayList<PartitionOffset>();
            partitionOffsets.add(insertChunk.getPartitionOffset());

            int batchSize = properties.getBatchSize();
            while (!insertChunkQueue.isEmpty()) {
                val chunk = insertChunkQueue.poll();
                if (chunk != null) {
                    insertChunk.getInsertSqlRequest().getParams()
                            .addAll(chunk.getInsertSqlRequest().getParams());
                    partitionOffsets.add(chunk.getPartitionOffset());
                    if (--batchSize == 0) break;
                } else {
                    break;
                }
            }

            executor.processChunk(insertChunk)
                    .map(rows -> {
                        publish(partitionOffsets);
                        return rows;
                    })
                    .onSuccess(rows -> log.debug("Written lines [{}] to data source", insertChunk.getInsertSqlRequest().getParams().size()))
                    .onFailure(t -> log.error("Exception during publishing offsets for commit: [{}]", partitionOffsets, t))
                    .<Void>mapEmpty()
                    .onComplete(promise);
        });
    }

    @SneakyThrows
    private void publish(ArrayList<PartitionOffset> partitionOffsets) {
        vertx.eventBus().publish(KafkaCommitVerticle.KAFKA_COMMIT_TOPIC + context.getId(),
                DatabindCodec.mapper().writeValueAsString(partitionOffsets));
    }

    private void error(InsertDataContext context, Throwable throwable) {
        val contextId = context.getId();
        context.setCause(throwable);
        log.error("Error consuming message in context: {}", context, throwable);
        stop(contextId);
    }

    private void stop(InsertDataContextId insertDataContextId) {
        vertx.eventBus().publish(DataTopic.SEND_RESPONSE.getValue(), insertDataContextId.getId());
    }

    @Override
    public DeploymentOptions getDeploymentOptions() {
        return new DeploymentOptions()
                .setWorker(true)
                .setWorkerPoolName("insert-worker-pool")
                .setWorkerPoolSize(properties.getPoolSize());
    }

}
