/*
 * Copyright © 2022 DATAMART LLC
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ru.datamart.kafka.postgres.writer.verticle;

import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.core.type.TypeReference;
import io.vertx.core.CompositeFuture;
import io.vertx.core.DeploymentOptions;
import io.vertx.core.Future;
import io.vertx.core.Promise;
import io.vertx.core.eventbus.Message;
import io.vertx.core.json.jackson.DatabindCodec;
import io.vertx.kafka.client.common.TopicPartition;
import lombok.Builder;
import lombok.Getter;
import lombok.extern.slf4j.Slf4j;
import lombok.val;
import ru.datamart.kafka.postgres.writer.configuration.properties.VerticleProperties;
import ru.datamart.kafka.postgres.writer.factory.InsertRequestFactory;
import ru.datamart.kafka.postgres.writer.model.DataTopic;
import ru.datamart.kafka.postgres.writer.model.InsertDataContext;
import ru.datamart.kafka.postgres.writer.model.InsertDataContextId;
import ru.datamart.kafka.postgres.writer.model.kafka.PartitionOffset;
import ru.datamart.kafka.postgres.writer.model.kafka.TopicPartitionConsumer;
import ru.datamart.kafka.postgres.writer.service.kafka.KafkaConsumerService;

import java.util.*;
import java.util.concurrent.ConcurrentLinkedDeque;
import java.util.stream.Collectors;


@Slf4j
@Getter
@Builder
public class KafkaCommitVerticle extends ConfigurableVerticle {
    public static final String START_COMMIT = "kafka_commit_start";
    public static final String KAFKA_COMMIT_TOPIC = "kafka_consumer_offset_commit";
    private final VerticleProperties.CommitWorkerProperties workerProperties;
    private final HashMap<TopicPartition, TopicPartitionConsumer> consumerMap;
    private final InsertRequestFactory insertRequestFactory;
    private final String id = UUID.randomUUID().toString();
    private final Queue<PartitionOffset> offsetsQueue = new ConcurrentLinkedDeque<>();
    private final KafkaConsumerService consumerService;
    private final InsertDataContext context;
    private volatile long timerId;
    private volatile boolean stopped;

    @Override
    public void start() {
        vertx.eventBus().consumer(START_COMMIT + context.getId(), ar -> startTimer());
        vertx.eventBus().consumer(KAFKA_COMMIT_TOPIC + context.getId(), this::collectKafkaMessages);
    }

    @Override
    public void stop(Promise<Void> promise) {
        stopped = true;
        vertx.cancelTimer(timerId);
        commitOffsets().onComplete(ar -> {
            if (ar.failed()) {
                log.error("Could not commit offset on undeploy", ar.cause());
            }
            promise.complete();
        });
    }

    private void collectKafkaMessages(Message<Object> partitionOffsets) {
        try {
            val offsets = DatabindCodec.mapper()
                    .readValue(partitionOffsets.body().toString(), new TypeReference<List<PartitionOffset>>() {
                    });
            offsetsQueue.addAll(offsets);
        } catch (JsonProcessingException e) {
            log.error("Deserialize PartitionOffsets error [{}]: {}", partitionOffsets, e);
            error(context, e);
        }
    }

    private void startTimer() {
        if (stopped) {
            return;
        }

        timerId = vertx.setTimer(workerProperties.getCommitPeriodMs(), timer -> {
            if (stopped) {
                return;
            }

            commitOffsets()
                    .onSuccess(v -> startTimer())
                    .onFailure(t -> error(context, t));
        });
    }

    private Future<Void> commitOffsets() {
        return Future.future(promise -> {
            val partitionOffsets = new ArrayList<PartitionOffset>();
            while (!offsetsQueue.isEmpty()) {
                val offset = offsetsQueue.poll();
                if (offset != null) {
                    partitionOffsets.add(offset);
                } else {
                    break;
                }
            }

            if (partitionOffsets.isEmpty()) {
                promise.complete();
                return;
            }

            commitKafkaMessages(partitionOffsets)
                    .map(notExistOffsets -> {
                        offsetsQueue.addAll(notExistOffsets);
                        return notExistOffsets;
                    })
                    .<Void>mapEmpty()
                    .onComplete(promise);
        });
    }

    private Future<List<PartitionOffset>> commitKafkaMessages(List<PartitionOffset> partitionOffsets) {
        return Future.future(promise -> {
            val partitionListMap = partitionOffsets.stream()
                    .collect(Collectors.groupingBy(PartitionOffset::getPartition, Collectors.toList()));

            val notExistsConsumers = partitionListMap.entrySet().stream()
                    .filter(e -> !consumerMap.containsKey(e.getKey()))
                    .flatMap(e -> e.getValue().stream())
                    .collect(Collectors.toList());

            val commitFutures = new ArrayList<Future>();
            consumerMap.values().forEach(topicPartitionConsumer -> {
                val offsetsToCommit = partitionListMap.get(topicPartitionConsumer.getTopicPartition());
                if (offsetsToCommit == null) {
                    return;
                }
                val committedChunks = offsetsToCommit.size();

                offsetsToCommit.forEach(partitionOffset -> topicPartitionConsumer.addCompletedOffset(partitionOffset.getOffset()));
                commitFutures.add(
                        topicPartitionConsumer.commitCompletedOffsets()
                                .onSuccess(ignored -> context.getActiveChunks().addAndGet(-committedChunks))
                );
            });

            CompositeFuture.join(commitFutures)
                    .map(ignored -> {
                        consumerMap.values().forEach(topicPartitionConsumer -> topicPartitionConsumer.getKafkaConsumer().resume());
                        return notExistsConsumers;
                    })
                    .onComplete(promise);
        });
    }

    private void error(InsertDataContext context, Throwable throwable) {
        val contextId = context.getId();
        context.setCause(throwable);
        log.error("Error consuming message in context: {}", context, throwable);
        stop(contextId);
    }

    private void stop(InsertDataContextId insertDataContextId) {
        vertx.eventBus().publish(DataTopic.SEND_RESPONSE.getValue(), insertDataContextId.getId());
    }

    @Override
    public DeploymentOptions getDeploymentOptions() {
        return new DeploymentOptions()
                .setWorker(true)
                .setWorkerPoolName("kafka-commit-worker-pool")
                .setWorkerPoolSize(workerProperties.getPoolSize());
    }
}
