/*
 * Copyright © 2022 DATAMART LLC
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ru.datamart.kafka.postgres.writer.verticle;

import io.vertx.core.DeploymentOptions;
import io.vertx.kafka.client.common.PartitionInfo;
import io.vertx.kafka.client.common.TopicPartition;
import io.vertx.kafka.client.consumer.KafkaConsumerRecord;
import lombok.Builder;
import lombok.Getter;
import lombok.extern.slf4j.Slf4j;
import lombok.val;
import ru.datamart.kafka.postgres.avro.codec.AvroDecoder;
import ru.datamart.kafka.postgres.writer.configuration.properties.VerticleProperties;
import ru.datamart.kafka.postgres.writer.factory.InsertRequestFactory;
import ru.datamart.kafka.postgres.writer.model.DataTopic;
import ru.datamart.kafka.postgres.writer.model.InsertDataContext;
import ru.datamart.kafka.postgres.writer.model.kafka.Chunk;
import ru.datamart.kafka.postgres.writer.model.kafka.InsertChunk;
import ru.datamart.kafka.postgres.writer.model.kafka.PartitionOffset;
import ru.datamart.kafka.postgres.writer.model.kafka.TopicPartitionConsumer;
import ru.datamart.kafka.postgres.writer.service.kafka.KafkaConsumerService;

import java.util.Collections;
import java.util.HashMap;
import java.util.Queue;


@Slf4j
@Getter
@Builder
public class KafkaConsumerVerticle extends ConfigurableVerticle {
    public static final String START_TOPIC = "kafka_consumer_start";
    public static final int DELAY = 500;
    private final AvroDecoder decoder = new AvroDecoder();
    private final VerticleProperties.KafkaConsumerWorkerProperties workerProperties;
    private final InsertRequestFactory insertRequestFactory;
    private final KafkaConsumerService consumerService;
    private final PartitionInfo partitionInfo;
    private final InsertDataContext context;
    private final Queue<InsertChunk> insertChunkQueue;
    private final HashMap<TopicPartition, TopicPartitionConsumer> consumerMap;
    private volatile TopicPartitionConsumer topicPartitionConsumer;
    private volatile boolean stopped;

    @Override
    public void start() {
        vertx.eventBus().consumer(START_TOPIC + context.getId(), ar -> startConsumerLogic());
    }

    @Override
    public void stop() {
        stopped = true;
        if (topicPartitionConsumer != null) {
            topicPartitionConsumer.getKafkaConsumer().handler(null);
            topicPartitionConsumer.getKafkaConsumer().close();
        }
    }

    private void startConsumerLogic() {
        consumerService.createTopicPartitionConsumer(context, partitionInfo)
                .onSuccess(consumer -> {
                    if (stopped) {
                        consumer.getKafkaConsumer().close();
                        return;
                    }

                    setConsumer(consumer);
                    consumer.getKafkaConsumer().handler(this::processRecord);
                });
    }

    private void processRecord(KafkaConsumerRecord<byte[], byte[]> record) {
        if (context.getCause() != null) {
            return;
        }

        int activeChunks = context.getActiveChunks().incrementAndGet();
        if (activeChunks >= workerProperties.getMaxFetchSize()) {
            topicPartitionConsumer.getKafkaConsumer().pause();
        }

        try {
            val chunk = parseChunk(record);
            val postgresInsertSqlRequest = insertRequestFactory.create(context, chunk.getRows());
            val insertChunk = new InsertChunk(new PartitionOffset(
                    chunk.getTopicPartition(),
                    chunk.getOffset()),
                    postgresInsertSqlRequest);
            insertChunkQueue.add(insertChunk);
            log.trace("Received chunk for recording: [{}]", chunk);
        } catch (Exception e) {
            stopConsuming(context, e);
        }
    }

    private Chunk parseChunk(KafkaConsumerRecord<byte[], byte[]> record) {
        if (record.value() == null || record.value().length == 0) {
            return Chunk.builder()
                    .topicPartition(new TopicPartition(record.topic(), record.partition()))
                    .offset(record.offset())
                    .rows(Collections.emptyList())
                    .build();
        }

        return Chunk.builder()
                .topicPartition(new TopicPartition(record.topic(), record.partition()))
                .offset(record.offset())
                .rows(decoder.decode(record.value()))
                .build();
    }

    private void setConsumer(TopicPartitionConsumer consumer) {
        topicPartitionConsumer = consumer;
        consumerMap.put(consumer.getTopicPartition(), consumer);
    }

    private void stopConsuming(InsertDataContext context, Throwable throwable) {
        val insertDataContextId = context.getId();
        context.setCause(throwable);
        log.error("Error consuming message in context: {}", context, throwable);
        vertx.eventBus().publish(DataTopic.SEND_RESPONSE.getValue(), insertDataContextId.getId());
    }

    @Override
    public DeploymentOptions getDeploymentOptions() {
        return new DeploymentOptions()
                .setWorker(true)
                .setWorkerPoolName("kafka-consumer-worker-pool")
                .setWorkerPoolSize(workerProperties.getPoolSize());
    }
}
